#include <complex.h>
#include <math.h>
#include <omp.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// CLASS = C
const int NX = 512;
const int NY = 512;
const int NZ = 512;
const int MAXDIM = 512;
const int NITER_DEFAULT = 20;

// total number of grid points with padding
const int NXP = NX + 1;
const int NTOTALP = NXP * NY * NZ;
const double NTOTAL_F = (double)NX * NY * NZ;

// If processor array is 1x1 -> 0D grid decomposition

// Cache blocking params. These values are good for most RISC processors.
// FFT parameters:
//   fftblock controls how many ffts are done at a time.
//   The default is appropriate for most cache-based machines
//   On vector machines, the FFT can be vectorized with vector
//   length equal to the block size, so the block size should
//   be as large as possible. This is the size of the smallest
//   dimension of the problem: 128 for class A, 256 for class B and
//   512 for class C.

const int FFTBLOCK_DEFAULT = 32;
const int FFTBLOCKPAD_DEFAULT = FFTBLOCK_DEFAULT + 2;

// other stuff
const int SEED = 314159265;
const double ALPHA = 1E-6;

bool debug;

int niter;
int fftblock;
int fftblockpad;

double complex *u;
double complex *u0;
double complex *u1;
double *twiddle;

int dims[3];

#define ix2d(i, d1, j, d2) ((j) * (d1)) + (i)
#define ix3d(i, d1, j, d2, k, d3) ((k) * (d2) * (d1)) + ix2d(i, d1, j, d2)
#define ix2dT(i, d1, j, d2) ix3d(i, d1, j, d2, omp_get_thread_num(), omp_get_max_threads())

static int ilog2(const int n)
{
	if (n == 1)
		return 0;

	int lg = 1;
	int nn = 2;

	while (nn < n) {
		nn *= 2;
		lg += 1;
	}

	return lg;
}

static void alloc_space(void)
{
	u = malloc(NXP * sizeof(*u));
	u0 = malloc(NTOTALP * sizeof(*u0));
	u1 = malloc(NTOTALP * sizeof(*u1));
	twiddle = malloc(NTOTALP * sizeof(*twiddle));

	if (u0 == NULL || u1 == NULL || twiddle == NULL) {
		printf("Error encountered in allocating space\n");
		exit(EXIT_FAILURE);
	}
}

static void setup(void)
{
	debug = false;
	niter = NITER_DEFAULT;

	printf("\n");
	printf("\n");
	printf(" NAS Parallel Benchmarks (NPB3.4-OMP) - FT Benchmark\n");
	printf("\n");
	printf(" Size                        : %dx%dx%d\n", NX, NY, NZ);
	printf(" Iterations                  : %d\n", niter);
	printf(" Number of available threads : %d\n", omp_get_max_threads());
	printf("\n");

	dims[0] = NX;
	dims[1] = NY;
	dims[2] = NZ;

	// ---------------------------------------------------------------------
	// Set up info for blocking of ffts and transposes.  This improves
	// performance on cache-based systems. Blocking involves
	// working on a chunk of the problem at a time, taking chunks
	// along the first, second, or third dimension.
	//
	// - In cffts1 blocking is on 2nd dimension (with fft on 1st dim)
	// - In cffts2/3 blocking is on 1st dimension (with fft on 2nd and 3rd dims)
	//
	// Since 1st dim is always in processor, we'll assume it's long enough
	// (default blocking factor is 16 so min size for 1st dim is 16)
	// The only case we have to worry about is cffts1 in a 2d decomposition.
	// so the blocking factor should not be larger than the 2nd dimension.
	// ---------------------------------------------------------------------

	fftblock = FFTBLOCK_DEFAULT;
	fftblockpad = FFTBLOCKPAD_DEFAULT;

	if (fftblock != FFTBLOCK_DEFAULT)
		fftblockpad = fftblock + 3;
}

static void init_ui(double complex *u0,
                    double complex *u1,
                    double *twiddle,
                    const int d1,
                    const int d2,
                    const int d3)
{
	#pragma omp parallel for collapse(2)
	for (int k = 0; k < d3; k++) {
		for (int j = 0; j < d2; j++) {
			for (int i = 0; i < d1 + 1; i++) {
				#define index ix3d(i, d1 + 1, j, d2, k, d3)

				u0[index] = 0;
				u1[index] = 0;
				twiddle[index] = 0;

				#undef index
			}
		}
	}
}

// ---------------------------------------------------------------------
//  compute function from local (i,j,k) to ibar^2+jbar^2+kbar^2
//  for time evolution exponent.
// ---------------------------------------------------------------------
static void compute_indexmap(double *twiddle, const int d1, const int d2, const int d3)
{
	const double AP = -4 * ALPHA * M_PI * M_PI;

	// ---------------------------------------------------------------------
	// basically we want to convert the fortran indices
	//   1 2 3 4 5 6 7 8
	// to
	//   0 1 2 3 -4 -3 -2 -1
	// The following magic formula does the trick:
	// mod(i-1+n/2, n) - n/2
	// ---------------------------------------------------------------------

	#pragma omp parallel for collapse(2)
	for (int k = 0; k < d3; k++) {
		for (int j = 0; j < d2; j++) {
			const int kk = (k + d3 / 2) % d3 - d3 / 2;
			const int kk2 = kk * kk;

			const int jj = (j + d2 / 2) % d2 - d2 / 2;
			const int kj2 = jj * jj + kk2;

			for (int i = 0; i < d1; i++) {
				#define index ix3d(i, d1 + 1, j, d2, k, d3)

				const int ii = (i + d1 / 2) % d1 - d1 / 2;
				twiddle[index] = exp(AP * (ii * ii + kj2));

				#undef index
			}
		}
	}
}

static void compute_initial_conditions(double complex *u0, const int d1, const int d2, const int d3)
{
	srand48(SEED);

	for (int k = 0; k < d3; k++) {
		for (int j = 0; j < d2; j++) {
			for (int i = 0; i < d1 + 1; i++) {
				#define index ix3d(i, d1 + 1, j, d2, k, d3)

				u0[index] = drand48();

				#undef index
			}
		}
	}
}

static void evolve(double complex *u0,
                   double complex *u1,
                   double *twiddle,
                   const int d1,
                   const int d2,
                   const int d3)
{
	// ---------------------------------------------------------------------
	//  evolve u0 -> u1 (t time steps) in fourier space
	// ---------------------------------------------------------------------

	#pragma omp parallel for collapse(2)
	for (int k = 0; k < d3; k++) {
		for (int j = 0; j < d2; j++) {
			for (int i = 0; i < d1; i++) {
				#define index ix3d(i, d1 + 1, j, d2, k, d3)

				u0[index] *= twiddle[index];
				u1[index] = u0[index];

				#undef index
			}
		}
	}
}

// ---------------------------------------------------------------------
//  compute the roots-of-unity array that will be used for subsequent FFTs.
// ---------------------------------------------------------------------
static void fft_init(int n)
{
	// ---------------------------------------------------------------------
	//   Initialize the U array with sines and cosines in a manner that permits
	//   stride one access at each FFT iteration.
	// ---------------------------------------------------------------------

	const int m = ilog2(n);
	u[0] = m;

	int ku = 1;
	int ln = 1;

	for (int j = 0; j < m; j++) {
		const double t = M_PI / ln;

		for (int i = 0; i < ln; i++) {
			const double ti = i * t;

			u[i + ku] = cos(ti) + I * sin(ti);
		}

		ku += ln;
		ln *= 2;
	}
}

static void fftz2(const int is,
                  const int l,
                  const int m,
                  const int n,
                  const int ny,
                  const int ny1,
                  double complex *u,
                  double complex *x,
                  double complex *y)
{
	// ---------------------------------------------------------------------
	//    Performs the L-th iteration of the second variant of the Stockham FFT.
	// --------------------------------------------------------------------

	const int n1 = n / 2;
	const int lk = 1 << (l - 1);
	const int li = 1 << (m - l);
	const int lj = 2 * lk;

	for (int i = 0; i < li; i++) {
		const int i11 = i * lk;
		const int i12 = i11 + n1;
		const int i21 = i * lj;
		const int i22 = i21 + lk;

		double complex u1;

		if (is >= 1)
			u1 = u[li + i];
		else
			u1 = conj(u[li + i]);

		// ---------------------------------------------------------------------
		//    This loop is vectorizable.
		// ---------------------------------------------------------------------
		for (int k = 0; k < lk; k++) {
			for (int j = 0; j < ny; j++) {
				#define index11 ix2dT(j, ny1, i11 + k, n)
				#define index12 ix2dT(j, ny1, i12 + k, n)

				const double complex x11 = x[index11];
				const double complex x21 = x[index12];

				#undef index21
				#undef index22

				#define index21 ix2dT(j, ny1, i21 + k, n)
				#define index22 ix2dT(j, ny1, i22 + k, n)

				y[index21] = x11 + x21;
				y[index22] = u1 * (x11 - x21);

				#undef index11
				#undef index12
			}
		}
	}
}

static void cfftz(const int is, const int m, const int n, double complex *x, double complex *y)
{
	// ---------------------------------------------------------------------
	//    Computes NY N-point complex-to-complex FFTs of X using an algorithm due
	//    to Swarztrauber.  X is both the input and the output array, while Y is a
	//    scratch array.  It is assumed that N = 2^M.  Before calling CFFTZ to
	//    perform FFTs, the array U must be initialized by calling CFFTZ with IS
	//    set to 0 and M set to MX, where MX is the maximum value of M for any
	//    subsequent call.
	// ---------------------------------------------------------------------

	// ---------------------------------------------------------------------
	//    Check if input parameters are invalid.
	// ---------------------------------------------------------------------

	const int mx = u[0];

	if ((is != 1 && is != -1) || m < 1 || m > mx) {
		// clang-format off
		printf("CFFTZ: Either U has not been initialized, or else one "
		       "of the input parameters is invalid %d %d %d\n", is, m, mx);
		// clang-format on
		exit(EXIT_FAILURE);
	}

	// ---------------------------------------------------------------------
	//    Perform one variant of the Stockham FFT.
	// ---------------------------------------------------------------------

	for (int l = 1; l <= m; l += 2) {
		fftz2(is, l, m, n, fftblock, fftblockpad, u, x, y);

		if (l == m) {
			// ---------------------------------------------------------------------
			//   Copy Y to X.
			// ---------------------------------------------------------------------

			for (int j = 0; j < n; j++) {
				for (int i = 0; i < fftblock; i++) {
					#define index ix2dT(i, fftblockpad, j, n)

					x[index] = y[index];

					#undef index
				}
			}

			continue;
		}

		fftz2(is, l + 1, m, n, fftblock, fftblockpad, u, y, x);
	}
}

static void cffts1(const int is,
                   const int d1,
                   const int d2,
                   const int d3,
                   const double complex *x,
                   double complex *xout,
                   double complex *y1,
                   double complex *y2)
{
	const int logd1 = ilog2(d1);

	#pragma omp parallel for collapse(2)
	for (int k = 0; k < d3; k++) {
		for (int jn = 0; jn < d2 / fftblock; jn++) {
			const int jj = jn * fftblock;

			for (int j = 0; j < fftblock; j++) {
				for (int i = 0; i < d1; i++) {
					#define indexY ix2dT(j, fftblockpad, i, d1)
					#define indexX ix3d(i, d1 + 1, j + jj, d2, k, d3)

					y1[indexY] = x[indexX];

					#undef indexY
					#undef indexX
				}
			}

			cfftz(is, logd1, d1, y1, y2);

			for (int j = 0; j < fftblock; j++) {
				for (int i = 0; i < d1; i++) {
					#define indexY ix2dT(j, fftblockpad, i, d1)
					#define indexX ix3d(i, d1 + 1, j + jj, d2, k, d3)

					xout[indexX] = y1[indexY];

					#undef indexY
					#undef indexX
				}
			}
		}
	}
}

void cffts2(const int is,
            const int d1,
            const int d2,
            const int d3,
            const double complex *x,
            double complex *xout,
            double complex *y1,
            double complex *y2)
{
	const int logd2 = ilog2(d2);

	#pragma omp parallel for collapse(2)
	for (int k = 0; k < d3; k++) {
		for (int in = 0; in < d1 / fftblock; in++) {
			const int ii = in * fftblock;

			for (int j = 0; j < d2; j++) {
				for (int i = 0; i < fftblock; i++) {
					#define indexY ix2dT(i, fftblockpad, j, d2)
					#define indexX ix3d(i + ii, d1 + 1, j, d2, k, d3)

					y1[indexY] = x[indexX];

					#undef indexY
					#undef indexX
				}
			}

			cfftz(is, logd2, d2, y1, y2);

			for (int j = 0; j < d2; j++) {
				for (int i = 0; i < fftblock; i++) {
					#define indexY ix2dT(i, fftblockpad, j, d2)
					#define indexX ix3d(i + ii, d1 + 1, j, d2, k, d3)

					xout[indexX] = y1[indexY];

					#undef indexY
					#undef indexX
				}
			}
		}
	}
}

static void cffts3(const int is,
                   const int d1,
                   const int d2,
                   const int d3,
                   const double complex *x,
                   double complex *xout,
                   double complex *y1,
                   double complex *y2)
{
	const int logd3 = ilog2(d3);

	#pragma omp parallel for collapse(2)
	for (int j = 0; j < d2; j++) {
		for (int in = 0; in < d1 / fftblock; in++) {
			const int ii = in * fftblock;

			for (int k = 0; k < d3; k++) {
				for (int i = 0; i < fftblock; i++) {
					#define indexY ix2dT(i, fftblockpad, k, d3)
					#define indexX ix3d(i + ii, d1 + 1, j, d2, k, d3)

					y1[indexY] = x[indexX];

					#undef indexY
					#undef indexX
				}
			}

			cfftz(is, logd3, d3, y1, y2);

			for (int k = 0; k < d3; k++) {
				for (int i = 0; i < fftblock; i++) {
					#define indexY ix2dT(i, fftblockpad, k, d3)
					#define indexX ix3d(i + ii, d1 + 1, j, d2, k, d3)

					xout[indexX] = y1[indexY];

					#undef indexY
					#undef indexX
				}
			}
		}
	}
}

void fft(const int dir, double complex *x1, double complex *x2)
{
	double complex y1[fftblockpad * MAXDIM * omp_get_max_threads()];
	double complex y2[fftblockpad * MAXDIM * omp_get_max_threads()];

	// ---------------------------------------------------------------------
	//  note: args x1, x2 must be different arrays
	//  note: args for cfftsx are (direction, layout, xin, xout, scratch)
	//        xin/xout may be the same and it can be somewhat faster
	//        if they are
	// ---------------------------------------------------------------------

	if (dir == 1) {
		cffts1(1, dims[0], dims[1], dims[2], x1, x1, y1, y2);
		cffts2(1, dims[0], dims[1], dims[2], x1, x1, y1, y2);
		cffts3(1, dims[0], dims[1], dims[2], x1, x2, y1, y2);
	} else {
		cffts3(-1, dims[0], dims[1], dims[2], x1, x1, y1, y2);
		cffts2(-1, dims[0], dims[1], dims[2], x1, x1, y1, y2);
		cffts1(-1, dims[0], dims[1], dims[2], x1, x2, y1, y2);
	}
}

static void checksum(const int i, double complex *u1, const int d1, const int d2, const int d3)
{
	double complex chk = 0;

	#pragma omp parallel for reduction(+ : chk)
	for (int j = 1; j <= 1024; j++) {
		const int q = j % NX;
		const int r = (3 * j) % NY;
		const int s = (5 * j) % NZ;

		#define index ix3d(q, d1 + 1, r, d2, s, d3)

		chk += u1[index];

		#undef index
	}

	chk /= NTOTAL_F;

	printf(" T = %d     Checksum = %.10E %.10E\n", i, __real__ chk, __imag__ chk);
}

int main(void)
{
	alloc_space();
	setup();
	init_ui(u0, u1, twiddle, dims[0], dims[1], dims[2]);

	// ---------------------------------------------------------------------
	// Run the entire problem once to make sure all data is touched.
	// This reduces variable startup costs, which is important for such a
	// short benchmark. The other NPB 2 implementations are similar.
	// ---------------------------------------------------------------------
	compute_indexmap(twiddle, dims[0], dims[1], dims[2]);
	compute_initial_conditions(u1, dims[0], dims[1], dims[2]);
	fft_init(dims[0]);
	fft(1, u1, u0);

	// ---------------------------------------------------------------------
	// Start over from the beginning. Note that all operations must
	// be timed, in contrast to other benchmarks.
	// ---------------------------------------------------------------------
	compute_indexmap(twiddle, dims[0], dims[1], dims[2]);
	compute_initial_conditions(u1, dims[0], dims[1], dims[2]);
	fft_init(dims[0]);
	fft(1, u1, u0);

	for (int iter = 1; iter <= niter; iter++) {
		evolve(u0, u1, twiddle, dims[0], dims[1], dims[2]);
		fft(-1, u1, u1);
		checksum(iter, u1, dims[0], dims[1], dims[2]);
	}

	return EXIT_SUCCESS;
}
